from datetime import datetime
from io import StringIO
from pathlib import Path
from typing import Mapping, Union

import pandas as pd
from tqdm import tqdm


def read_ls(file: Union[str, Path, StringIO]) -> pd.DataFrame:
    """
    Reads the output of an `ls -l`-like command and returns a DataFrame.
    Converts 'Month', 'Day', and 'Time' columns into a single 'Datetime' column.

    Parameters:
        file (str | Path | StringIO): Input file path or file-like object.

    Returns:
        pd.DataFrame: Parsed file content with a 'Datetime' column added.
    """
    original_columns = [
        "Permissions",
        "Hardlinks",
        "Owner",
        "Group",
        "Filesize",
        "Month",
        "Day",
        "Time",
        "Filename",
    ]

    def parse_dataframe(iterable) -> pd.DataFrame:
        df = pd.read_fwf(iterable, names=original_columns)

        # Combine Month, Day, and Time into a Datetime column
        current_year = datetime.now().year
        df["Datetime"] = pd.to_datetime(
            df["Month"] + " " + df["Day"].astype(str) + f" {current_year} " + df["Time"],
            format="%b %d %Y %H:%M",
            errors="coerce"
        )

        # Define new column order: same as original, but replacing date fields with 'Datetime'
        final_columns = [
            "Permissions",
            "Hardlinks",
            "Owner",
            "Group",
            "Filesize",
            "Datetime",
            "Filename",
        ]

        return df.drop(columns=["Month", "Day", "Time"])[final_columns].sort_values("Datetime", ignore_index=True)

    if isinstance(file, (str, Path)):
        with open(file, "r", encoding="utf-8") as f:
            next(f, None)  # Skip header line
            return parse_dataframe(f)

    elif isinstance(file, StringIO):
        return parse_dataframe(file)

    else:
        raise TypeError(f"'file' must be str, Path, or StringIO, not {type(file).__name__}")


def read_filetree_ls(file: str | Path):
    chunks = {}
    current_io = StringIO()
    current_name = None
    with open(file, "r", encoding="utf-8") as rf:
        for i, line in enumerate(rf):
            if line.strip() == "":
                if current_name is not None:
                    chunks[current_name] = current_io
                    current_name = None
                    current_io = StringIO()
                continue
            if line.startswith("====="):
                current_name = line.strip().strip("=").strip()
                next(rf)
                next(rf)
                continue
            current_io.write(line)
    
    # return chunks
    ls_dict = {}
    for fpath, io in tqdm(chunks.items()):
        try:
            ls_dict[fpath] = read_ls(StringIO(io.getvalue()))
        except Exception as exc:
            ls_dict[fpath] = f"Error: {str(exc)}"
            continue

    return ls_dict
